# Copyright 2016-2017 Christian Shtarkov
#
# This file is part of Autobump.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Provide functions that get configuration parameters.

Methods check these sources in succession and stop at the first one:

1. Environment variables.
2. Configuration file.
3. Defaults.

Additionally, configuration may be overriden at runtime by using
the context manager and decorator defined here.
"""

import os
import re
import logging
import configparser
from copy import deepcopy

logger = logging.getLogger(__name__)

CONFIG_FILE = "autobump.ini"
falsy = re.compile(r"^([Ff](alse)?|0+)$")
defaults = {
    "autobump": {
        "git": "git",
        "hg": "hg",
        "clojure": "clojure",
        "java": "java",
        "javac": "javac"
    },

    "only_consider": {
        "files": [],
        "dirs": [],
        "entities": []
    },

    "ignore": {
        "files": ["project.clj", "setup.py", "setupegg.py"],
        "dirs": [".git", ".hg", "test", "tests", "doc", "docs"],
        "entities": [],
    },

    "python": {
        "omit_on_error": False,
        "structural_typing": True,
        "type_hinting": True
    },

    "java_native": {
        "lazy_type_checking": False,
        "classpath": ""
    },

    "java_ast": {
        "error_on_external_types": True
    },

    "clojure": {
        "lazy_type_checking": True,
        "classpath": ""
    },
}
cached = dict()


def get(category, name):
    """Get the value of a configuration parameter
    by checking several sources in succession."""
    value = cached.get((category, name), None)
    if value is not None:
        return value

    # Check environment variables.
    env_name = "AB_" + name.upper()
    if env_name in os.environ:
        value = os.environ[env_name]

    # Check configuration file.
    else:
        config = configparser.ConfigParser()
        config.read(CONFIG_FILE)
        value = config.get(category, name, fallback=defaults[category][name])

    if falsy.match(str(value)):
        value = False

    logger.info("{}/{} is {}"
                .format(category, name, (value if value != "" else "not set")))

    cached[(category, name)] = value
    return value


def make_get(category, name):
    """Return a no-arguments function that
    gets the value of a parameter."""
    return lambda: get(category, name)


class config_overrides(object):
    """Context manager that temporary overrides
    the values of some parameters."""
    def __init__(self, overrides):
        self.overrides = overrides

    def __enter__(self):
        global cached
        self.previous = deepcopy(cached)
        for category in self.overrides:
            for name in self.overrides[category]:
                set(category, name, self.overrides[category][name])

    def __exit__(self, *args):
        global cached
        cached = deepcopy(self.previous)


def config_override(category, name, value):
    """Decorator that overrides the value of just
    one parameter."""
    def wrap(f):
        def wrapped(*args, **kwargs):
            with config_overrides({category: {name: value}}):
                return f(*args, **kwargs)
        return wrapped
    return wrap


def set(category, name, value):
    """Permanently override the value of an option."""
    global cached
    cached[(category, name)] = value


# autobump
# TODO: move these to respective handlers
# or rename this to "executables"
git = make_get("autobump", "git")
hg = make_get("autobump", "hg")
clojure = make_get("autobump", "clojure")
java = make_get("autobump", "java")
javac = make_get("autobump", "javac")


# ignore
def ignored(what, name):
    """Check whether something should be ignored."""
    only_consider_lit = get("only_consider", what)
    ignored_lit = get("ignore", what)
    if isinstance(only_consider_lit, str):
        only_consider_lit = only_consider_lit.splitlines()
    if isinstance(ignored_lit, str):
        ignored_lit = ignored_lit.splitlines()

    if len(only_consider_lit) > 0:
        return name not in only_consider_lit

    return name in ignored_lit


def file_ignored(name):
    """Check whether a file should be ignored."""
    return ignored("files", name)


def dir_ignored(name):
    """Check whether a directory should be ignored."""
    return ignored("dirs", name)


def entity_ignored(name):
    """Check whether an entity should be ignored."""
    return ignored("entities", name)


# python
python_omit_on_error = make_get("python", "omit_on_error")
structural_typing = make_get("python", "structural_typing")
type_hinting = make_get("python", "type_hinting")

# java_native
java_lazy_type_checking = make_get("java_native", "lazy_type_checking")


def java_classpath():
    # The classpath is special, because we need to append
    # the working directory so that Autobump's utilities like
    # the Inspector and TypeCompatibilityChecker work.
    value = get("java_native", "classpath")
    return (value + ":.:") if value != "" else ""

# java_ast
java_error_on_external_types = make_get("java_ast", "error_on_external_types")

# clojure
clojure_lazy_type_checking = make_get("clojure", "lazy_type_checking")
clojure_classpath = make_get("clojure", "classpath")



# Copyright 2016-2017 Christian Shtarkov
#
# This file is part of Autobump.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Common classes and functions used throughout Autobump.
"""

import re
import logging
import subprocess

from autobump import diff

logger = logging.getLogger(__name__)


class VersionControlException(Exception):
    pass


class Semver(object):
    """Minimal representation of a semantic version."""

    class NotAVersionNumber(Exception):
        pass

    def __init__(self, major, minor, patch):
        assert type(major) is int
        assert type(minor) is int
        assert type(patch) is int
        self.major, self.minor, self.patch = major, minor, patch

    def __eq__(self, other):
        assert isinstance(other, type(self))
        return self.major == other.major and \
               self.minor == other.minor and \
               self.patch == other.patch

    def __ne__(self, other):
        return not self.__eq__(other)

    @classmethod
    def from_string(semver, version):
        major, minor, patch = [int(c) for c in version.split(".")]
        return semver(major, minor, patch)

    @classmethod
    def from_tuple(semver, version):
        major, minor, patch = version
        return semver(major, minor, patch)

    @classmethod
    def guess_from_string(semver, string):
        """Guess a version number from a tag name. """
        match = re.match(r"(v|ver|version)?-?(\d+)\.?(\d+)?\.?(\d+)?", string)
        if match:
            major = int(match.group(2))
            minor = int(match.group(3) or 0)
            patch = int(match.group(4) or 0)
            guess = semver(major, minor, patch)
            logger.warning("Guessing version from string '{}': {}"
                           .format(string, guess))
            return guess
        else:
            raise semver.NotAVersionNumber("Cannot reliable guess version number from {}".format(string))

    def bump(self, bump):
        """Bump version using a Bump enum."""
        assert type(bump) is diff.Bump, "Bump should be an Enum"
        if bump is diff.Bump.patch:
            return Semver(self.major, self.minor, self.patch + 1)
        if bump is diff.Bump.minor:
            return Semver(self.major, self.minor + 1, 0)
        if bump is diff.Bump.major:
            return Semver(self.major + 1, 0, 0)
        # No bump
        return Semver(self.major, self.minor, self.patch)

    def __str__(self):
        return str(self.major) + "." + str(self.minor) + "." + str(self.patch)


def popen(args, cwd="."):
    """Thinly wrap subprocess.Popen.

    Always pipes stdout and stderr so that nothing is seen by
    the end user unless explicitly printed.
    """
    child = subprocess.Popen(args,
                             cwd=cwd,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
    stdout_data, stderr_data = child.communicate()
    return (child.returncode,
            stdout_data.decode("ascii").strip(),
            stderr_data.decode("ascii").strip())



# Copyright 2016-2017 Christian Shtarkov
#
# This file is part of Autobump.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Convert a Clojure codebase into a list of Units.

The Clojure handler invokes a script which is inspected
to run in the same environment as the project. This is best achieved
by configuring the Clojure executable to be 'lein exec' or something similar.
"""

import os
import string
import logging

from autobump import config
from autobump.capir import Type, Field, Parameter, Signature, Function, Unit
from autobump.common import popen

logger = logging.getLogger(__name__)
libexec = os.path.join(os.path.dirname(__file__), "..", "libexec")
inspector_clj = os.path.join(libexec, "evalinspector.clj")
_source_file_ext = ".clj"


class _ClojureType(Type):
    def __init__(self, name):
        self.name = name

    def is_compatible(self, other):
        if self.name == "nil":
            return True
        if config.clojure_lazy_type_checking():
            return self.name == other.name
        else:
            raise NotImplementedError("Non-lazy type checking")

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<ClojureType {}>".format(self.name)


class _ClojureUtilityException(Exception):
    pass


class _SexpReadException(Exception):
    pass


def _run_inspector(files, repo):
    """Runs the utility program inspector.clj with a list of files, with
    the working directory set to 'repo'."""
    arglist = [config.clojure(), inspector_clj] + files
    logger.debug("Running inspector as follows: " + ' '.join(arglist))
    return_code, stdout, stderr = popen(arglist, cwd=repo)
    if return_code != 0:
        raise _ClojureUtilityException(stderr)
    return stdout


def _lookup_type(name):
    # TODO: Properly lookup types.
    return _ClojureType(name)


def _sexp_read(s):
    """Reads in a sexp describing a Clojure codebase
    and convert it into the common representation."""
    def tokenize(s):
        delimiters = string.whitespace + "()"
        token = ""
        for char in s:
            if char in delimiters:
                if token != "":
                    yield token
                if char in "()":
                    yield char
                token = ""
            else:
                token = token + char
        if token != "":
            yield token

    def sexp_to_list(gen):
        lst = []
        for token in gen:
            if token == "(":
                lst.append(sexp_to_list(gen))
            elif token == ")":
                return lst
            else:
                lst.append(token)
        return lst

    def verify_tag(tag, lst):
        if lst[0] != tag:
            raise _SexpReadException("Expected {}, got {}".format(tag, lst[0]))

    def read_signature(lst):
        verify_tag("signature", lst)
        tag, positional, optional = lst
        parameters = [Parameter(name, _lookup_type(type_name))
                      for name, type_name
                      in positional]
        # TODO: Properly handle default values (:or idiom).
        parameters += [Parameter(name, _lookup_type(type_name), default_value=True)
                       for name, type_name
                       in optional]
        return Signature(parameters)

    def read_function(lst):
        verify_tag("function", lst)
        tag, name, signatures = lst
        return Function(name, _lookup_type("nil"), [read_signature(s) for s in signatures])

    def read_field(lst):
        verify_tag("field", lst)
        tag, name, type_name = lst
        return Field(name, _lookup_type(type_name))

    def read_file(lst):
        verify_tag("file", lst)
        tag, ns, fields, functions = lst
        fields_dict = {}
        functions_dict = {}
        for f in fields:
            field = read_field(f)
            fields_dict[field.name] = field
        for f in functions:
            function = read_function(f)
            functions_dict[function.name] = function
        return Unit(ns,
                    fields_dict,
                    functions_dict,
                    dict())

    # TODO: Why [0]?
    units = dict()
    lst = sexp_to_list(tokenize(s))
    if len(lst) > 1:
        raise _SexpReadException("Sexp contains more than one top-level form")
    files = lst[0]
    for f in files:
        unit = read_file(f)
        units[unit.name] = unit
    return units


def clojure_codebase_to_units(location):
    """Returns a list of Units representing a Clojure codebase in 'location'."""
    # Resolve classpath
    if "CLASSPATH" in os.environ:
        logger.debug("CLASSPATH variable set (it's not guaranteed Clojure will it)")
    if config.clojure_classpath() != "":
        logger.debug("clojure/classpath set, setting that as CLASSPATH (it's not guaranteed Clojure will use it")
        os.environ["CLASSPATH"] = config.clojure_classpath()
    if "CLASSPATH" not in os.environ:
        logger.warning("No CLASSPATH set")
    else:
        logger.info("CLASSPATH is:\n\t{}".format(os.environ["CLASSPATH"]))

    # Inspect .clj files
    cljfiles = []
    for root, dirs, files in os.walk(location):
        dirs[:] = [d for d in dirs if not config.dir_ignored(d)]
        cljfiles += [os.path.join(root, f)
                     for f in files
                     if f.endswith(_source_file_ext) and not config.file_ignored(f)]

    return _sexp_read(_run_inspector(cljfiles, location))

build_required = False
codebase_to_units = clojure_codebase_to_units



# Copyright 2016-2017 Christian Shtarkov
#
# This file is part of Autobump.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Implement source control handling for Mercurial.
"""

import os
import tempfile

from autobump import config
from autobump.common import popen, VersionControlException


def _clone_repo(repo, checkout_dir):
    """Clone a hg repository into a directory."""
    return_code, _, _ = popen([config.hg(), "clone", repo, checkout_dir])
    if return_code != 0:
        raise VersionControlException("Cloning {} into {} failed!"
                                      .format(repo, checkout_dir))


def _checkout_commit(checkout_dir, commit):
    """Checkout a Hg commit at some location."""
    return_code, _, _ = popen([config.hg(), "update", commit], cwd=checkout_dir)
    if return_code != 0:
        raise VersionControlException("Checking out commit {} at {} failed!"
                                      .format(commit, checkout_dir))


def hg_get_commit(repo, commit):
    """Get a directory containing a commit found in a repository.

    The caller is responsible for cleaning up the directory afterwards
    by calling cleanup() on the handle."""
    repo_path = os.path.abspath(repo)
    repo_name = os.path.basename(repo)
    temp_dir_handle = tempfile.TemporaryDirectory()
    temp_dir = temp_dir_handle.name
    checkout_dir = os.path.join(temp_dir, repo_name)
    _clone_repo(repo_path, checkout_dir)
    _checkout_commit(checkout_dir, commit)
    return temp_dir_handle, checkout_dir


def hg_last_tag(repo):
    return_code, stdout, stderr = popen([config.hg(), "log", "-r", '"."', "--template", "{latesttag}"], cwd=repo)
    if return_code != 0:
        raise VersionControlException("Failed to get last tag of Hg repository {}"
                                      .format(repo))
    return stdout


def hg_all_tags(repo):
    return_code, stdout, stderr = popen([config.hg(), "log", "-r", 'tag()', "--template", '{tags}\n'], cwd=repo)
    if return_code != 0:
        raise VersionControlException("Failed to get tags of Hg repository {}"
                                      .format(repo))
    return stdout.split()


def hg_last_commit(repo):
    return_code, stdout, stderr = popen([config.hg(), "log", "-r", "tip", "--template", "{rev}"], cwd=repo)
    if return_code != 0:
        raise VersionControlException("Failed to get last commit of Hg repository {}"
                                             .format(repo))
    return stdout.split()[0]


get_commit = hg_get_commit
all_tags = hg_all_tags
last_tag = hg_last_tag
last_commit = hg_last_commit



# Copyright 2016-2017 Christian Shtarkov
#
# This file is part of Autobump.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Common API Representation (CAPIR)

This is a sufficiently abstract representation that models public
APIs found in most languages.
All language handlers eventually convert the public API
they found of a library to this representation.
This is the only thing that the diffing algorithm (diff.py)
can operate on.
"""

import uuid


class Entity(object):
    """Generic entity."""

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        raise TypeError("Comparing entity to something else.")

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return self.name


class Type(Entity):
    """Generic representation of a type.

    Language handlers are expected to use this type or derive from it.
    Most importantly, they should implement the 'is_compatible' method
    appropriately so that types can be compared to other types.
    """

    def __init__(self):
        self.name = str(uuid.uuid4())

    def is_compatible(self, other):
        """Checks whether 'self' can substitute 'other'."""
        return isinstance(self, type(other))

    def __eq__(self, other):
        return self.is_compatible(other) and self.name == other.name

    def __hash__(self):
        return hash(self.name)


class Field(Entity):
    """Top-level accessible field.

    This can be used to model module constants,
    class fields, and other similar public pieces of data."""

    def __init__(self, name, type):
        self.name = name
        self.type = type


class Parameter(Entity):
    """Parameter to a function.

    Has a type and default value, where the exact value of the
    default value is irrelevant: it only matters whether there is one.
    """

    def __init__(self, name, type, default_value=None):
        self.name = name
        self.type = type
        self.default_value = default_value

    def __hash__(self):
        return hash((self.name, self.type))

    def __str__(self):
        return "<{} {}>".format(self.type, self.name)

    def __repr__(self):
        return self.__str__()


class Signature(Entity):
    """Signature of a function.

    Essentially a collection of parameters. Functions can have
    multiple signatures to model things like overloading.
    """

    def __init__(self, parameters=None):
        if parameters is None:
            parameters = []
        self.parameters = parameters

    def __hash__(self):
        return hash(tuple(self.parameters))

    def __str__(self):
        return str(self.parameters)

    def __repr__(self):
        return self.__str__()


class Function(Entity):
    """Top-level function.

    Can be used to represent any callable piece of code,
    like functions, macros, methods and so on.
    """

    def __init__(self, name, type, signatures=None):
        self.name = name
        self.type = type
        self.signatures = signatures
        if self.signatures is None:
            self.signatures = []


class Unit(Entity):
    """Generic unit of code containing fields and functions.

    Could be a Java class, a Python module, a C translation unit and so on.
    """

    def __init__(self, name, fields, functions, units):
        self.name = name
        self.fields = fields
        self.functions = functions
        self.units = units



# Copyright 2016-2017 Christian Shtarkov
#
# This file is part of Autobump.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Convert a Java codebase into a list of Units, using introspection utilities written in Java.

This handler needs too additional Java utilities (found under 'libexec/') that do reflection
on compiled Java code.
"""

import os
import sys
import shutil
import logging
import tempfile
import subprocess
from xml.etree import ElementTree

from autobump import config
from autobump.capir import Type, Field, Parameter, Signature, Function, Unit
from autobump.common import popen

logger = logging.getLogger(__name__)
libexec = os.path.join(os.path.dirname(__file__), "..", "libexec")


# Type system
class _JavaNativeType(Type):
    """Representation of a Java type.

    When checking for compatibility with another type,
    an external utility will be invoked to perform introspection
    and find out if Java considers two types to be compatible."""
    def __init__(self, name, dimension=0):
        self.name = name
        self.dimension = dimension

    def is_compatible(self, other):
        assert type(self) is type(other), "Should never happen: comparing a _JavaNativeType to something else."
        assert hasattr(self, "location") and self.location is not None, "Should never happen: location should be set prior to calling is_compatible"
        if self.dimension != other.dimension:
            return False

        if config.java_lazy_type_checking():
            return self.name == other.name and self.dimension == other.dimension
        else:
            return _run_type_compatibility_checker(self.location, self.name, other.name)

    def __eq__(self, other):
        return self.is_compatible(other) and other.is_compatible(self)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<JavaNativeType {} {}>".format(self.dimension, self.name)

    def __hash__(self):
        return hash((self.dimension, self.name))


_dummyType = _JavaNativeType("dummy")
_dummyType.is_compatible = lambda t: True


class JavaUtilityException(Exception):
    pass


def _run_inspector(location, classnames):
    """Run the Inspector program to get the XML representation
    of classes found in 'location'."""
    return _run_utility("Inspector", [location] + classnames)


def _run_type_compatibility_checker(location, superclass, subclass):
    """Run the TypeCompatibilityChecker program and return a boolean
    indicating whether 'superclass' can really be substituted with 'subclass'."""
    output = _run_utility("TypeCompatibilityChecker", [location, superclass, subclass])
    return output == "true"


def _compile_and_run_utility(utility, args):
    """Compile a utility somewhere in a temporary directory
    so it can be used this time, and then run it."""

    logger.warning("Compiling {} in place".format(utility))
    filename = utility + ".java"
    # First, try to compile in place.
    return_code, stdout, stderr = popen([config.javac()] + [filename], cwd=libexec)
    if return_code != 0:
        logger.warning("Failed to compile {} in place, trying in a tempdir"
                       .format(utility))
        with tempfile.TemporaryDirectory() as dir:
            shutil.copy(os.path.join(libexec, filename), dir)
            return_code, stdout, stderr = popen([config.javac()] + [filename], cwd=dir)
            if return_code != 0:
                logger.error("Failed to compile {}! Please compile manually.".format(utility))
                raise JavaUtilityException("{} needs to be compiled".format(utility))
            # Call the utility from the temporary directory
            return _run_utility(utility, args, basedir=dir)
    # Compilation succeeded in-place, call the utility normally
    return _run_utility(utility, args)


def _run_utility(utility, args, basedir=libexec):
    """Run a Java utility program with arguments."""
    javafile = os.path.join(basedir, utility + ".java")
    classfile = os.path.join(basedir, utility + ".class")
    if os.path.isfile(javafile) and not os.path.isfile(classfile):
        logger.warning("{} has not been compiled".format(utility))
        return _compile_and_run_utility(utility, args)
    return_code, stdout, stderr = popen([config.java()] + [utility] + args, cwd=basedir)
    if return_code != 0:
        raise JavaUtilityException(stderr)
    return stdout


def _xml_element_to_type(elt):
    assert elt.tag == "type"
    return _JavaNativeType(elt.attrib["name"], int(elt.attrib["dimension"]))


def _xml_element_to_field(elt):
    """Convert an XML <field> into a Field."""
    assert elt.tag == "field"
    return Field(elt.attrib["name"], _xml_element_to_type(elt.find("type")))


def _xml_get_signature_of_method(elt):
    """Convert all <signature>s of a <method> into a Signature."""
    assert elt.tag == "method"
    signature_elt = elt.find("signature")
    parameter_elts = signature_elt.findall("parameter")
    parameters = [Parameter(p.attrib["name"], _xml_element_to_type(p.find("type")))
                  for p in parameter_elts]
    return_type = Parameter("$AUTOBUMP_RETURN$", _xml_element_to_type(elt.find("type")))
    parameters = [return_type] + parameters
    return Signature(parameters)


def _xml_element_to_unit(elt):
    """Convert an XML <class> element into a Unit."""
    functions = dict()
    fields = dict()
    units = dict()
    for child in elt:
        if child.tag == "field":
            field = _xml_element_to_field(child)
            fields[field.name] = field
        elif child.tag == "method":
            signature = _xml_get_signature_of_method(child)
            if child.attrib["name"] in functions:
                functions[child.attrib["name"]].signatures.append(signature)
            else:
                function = Function(child.attrib["name"], _dummyType, [signature])
                functions[function.name] = function
        elif child.tag == "class":
            unit = _xml_element_to_unit(child)
            units[unit.name] = unit
    return Unit(elt.attrib["name"], fields, functions, units)


def java_codebase_to_units(location, build_command, build_root):
    """Convert a Java codebase found at 'location' into a list of units.

    Works by compiling it with 'build_command' and then inspecting the
    class files under 'location/build_root'."""
    # Compile the classes
    logger.info("Starting build process")
    if "CLASSPATH" in os.environ:
        logger.debug("CLASSPATH variable set")
    if config.java_classpath() != "":
        logger.debug("java_native/classpath set, using that as classpath")
        os.environ["CLASSPATH"] = config.java_classpath()
    if "CLASSPATH" not in os.environ:
        logger.warning("No CLASSPATH set")
    else:
        logger.info("CLASSPATH is:\n\t{}".format(os.environ["CLASSPATH"]))

    try:
        subprocess.run(build_command,
                       cwd=location,
                       shell=True,
                       check=True,
                       stdout=sys.stderr,
                       stderr=sys.stderr)
    except subprocess.CalledProcessError:
        logger.error("Failed to call {}".format(build_command))
        exit(1)
    logger.info("Build completed")

    # Get absolute path to build root
    build_root = os.path.join(location, build_root)
    logger.debug("Absolute build root is {}".format(build_root))

    # Get a list of fully-qualified class names
    fqns = []
    for root, dirs, files in os.walk(build_root):
        dirs[:] = [d for d in dirs if not config.dir_ignored(d)]
        classfiles = [f for f in files if f.endswith(".class") and not config.file_ignored(f)]
        prefix = root[len(build_root):].replace(os.sep, ".")
        if len(prefix) > 0 and prefix[0] == ".":
            prefix = prefix[1:]
        fqns = fqns + [((prefix + ".") if prefix != "" else "") + os.path.splitext(n)[0] for n in classfiles]
    logger.debug("{} classes identified".format(len(fqns)))

    # Convert the XML representation of these classes to Unit
    xml = _run_inspector(build_root, fqns)
    root = ElementTree.fromstring(xml)
    units = dict()
    for child in root:
        # TODO: Validation of the XML shouldn't be done using assertions.
        # Is validation even necessary in this case?
        assert child.tag == "class", "Found element in XML that's not <class>!"
        unit = _xml_element_to_unit(child)
        units[unit.name] = unit
    return units

build_required = True
codebase_to_units = java_codebase_to_units



# Copyright 2016-2017 Christian Shtarkov
#
# This file is part of Autobump.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Tool for automatically suggesting the next version of a project
according to semantic versioning.

Autobump inspects how a version-controlled project has changed over time, identifies
the major, minor and patch changes as specified by semantic versioning, and proposes
a new version based on the previous one.

The tool works with multiple languages through the use of handlers (see 'handlers/')
by converting the public API of two revisions of a library into a common representation
(see 'capir.py') which can then be diffed (see 'diff.py') in order to get the new version.
"""

import os
import sys
import logging
import argparse
from functools import partial

from autobump import diff
from autobump.common import Semver, VersionControlException
from autobump.handlers import hg
from autobump.handlers import git
from autobump.handlers import python
from autobump.handlers import java_ast
from autobump.handlers import java_native
from autobump.handlers import clojure

logger = logging.getLogger(__name__)


def _identify_vcs(location):
    """Identify the type of a repository and return the appropriate handler."""
    try:
        # TODO: This will not work with bare repositories.
        if os.path.isdir(os.path.join(location, ".git")):
            return git
        elif os.path.isdir(os.path.join(location, ".hg")):
            return hg
        else:
            return None
    except IOError:
        logger.error("IO error occured while trying to get VCS")
        return None


def _patch_types_with_location(units, location):
    """Walk all types found in a dictionary of units and
    set their location property.

    This is necessary for some handlers, where the directory layout
    of the library matters.
    """
    for unit in units.values():
        for field in unit.fields.values():
            field.type.location = location
        for function in unit.functions.values():
            function.type.location = location
            for signature in function.signatures:
                for parameter in signature.parameters:
                    parameter.type.location = location
        _patch_types_with_location(unit.units, location)


def _evaluate(args, all_revisions):
    """Run Autobump in evaluation mode."""
    args.evaluate = False
    first_revision = args.f
    last_revision = args.to
    logger.info("Running in evaluation mode between {} and {}"
                .format(args.f, args.to))
    if first_revision not in all_revisions or last_revision not in all_revisions:
        logger.error("Invalid range, one or more tags not found!")
        exit(1)
    mismatched = []
    for rev_i in range(all_revisions.index(first_revision),
                       all_revisions.index(last_revision) - 1):
        a_revision = all_revisions[rev_i]
        b_revision = all_revisions[rev_i + 1]
        logger.debug("Evaluating revisions {}...{}"
                     .format(a_revision, b_revision))
        setattr(args, "from", a_revision)
        args.to = b_revision
        b_version_expected = Semver.guess_from_string(b_revision)
        b_version_actual = autobump(args=args)
        if b_version_expected != b_version_actual:
            logger.debug("Version found differs from name of tag:\n\tReported: {}\n\tFrom tag: {}"
                         .format(b_version_actual, b_version_expected))
            mismatched.append((a_revision, b_revision, b_version_actual))
    for a_revision, b_revision, actual in mismatched:
        print("{a_revision} --- {b_revision} should have been {a_revision} --- {actual}"
              .format(a_revision=a_revision, b_revision=b_revision, actual=actual))


def autobump(**kwargs):
    """Main entry point of Autobump."""
    description = """
Determine change of semantic version of code in a repository.

Example usage:

$ {0} python

    Shows what version the last commit should be, by guessing the
    previous version from the last tag.
    The tool will only look at Python files found in the repository.

$ {0} python --changelog changelog.txt

    Shows what version the last commit should be, by guessing the
    previous version from the last tag.
    Also, it records found changes to `changelog.txt`.
    The tool will only look at Python files found in the repository.

$ {0} python --from milestone-foo --from-version 1.1.0

    Shows what version the last commit should be, if the previous
    release was `milestone-foo` at version 1.1.0.
    `milestone-foo` can be a tag name or commit identifier.
    The tool will only look at Python files found in the repository.

$ {0} java --from milestone-foo --from-version 1.1.0 --to milestone-bar

    Shows what version `milestone-bar` should be, if the previous
    release was `milestone-foo` at version 1.1.0.
    The tool will only look at Java files found in the repository.
""".format(os.path.basename(sys.argv[0]))
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter, description=description)
    parser.add_argument("handler",
                        type=str,
                        help="what language handler to use {python, java_ast, java_native, clojure}")
    parser.add_argument("-c", "--changelog",
                        type=str,
                        help="generate changelog and write it to a file")
    parser.add_argument("-cstdout", "--changelog-stdout",
                        action="store_true",
                        help="write changelog to stdout (incompatible with `--changelog`)")
    parser.add_argument("-d", "--debug",
                        action="store_true",
                        help="print debugging information to stderr (implies `--info`)")
    parser.add_argument("-i", "--info",
                        action="store_true",
                        help="print progress information to stderr")
    parser.add_argument("-r", "--repo",
                        type=str,
                        help="repository location, will use working directory if not specified")
    parser.add_argument("-f", "--from",
                        type=str,
                        help="identifier of earlier revision, will use last tag if not specified")
    parser.add_argument("-fv", "--from-version",
                        type=str,
                        help="version of earlier revision, will try to guess if not specified!")
    parser.add_argument("-t", "--to",
                        type=str,
                        help="identifier of later revision, will use last commit if not specified")
    parser.add_argument("-bc", "--build-command",
                        type=str,
                        help="what shell command to run so that the project is built")
    parser.add_argument("-br", "--build-root",
                        type=str,
                        help="where the artifacts get placed after the project is built (relative to checkout location)")
    parser.add_argument("-e", "--evaluate",
                        action="store_true",
                        help="run in evaluation mode, measure quality of tags")

    # Parse command-line arguments only if not passed to the function.
    args = kwargs.get("args", parser.parse_args())
    args.f = getattr(args, "from")  # Syntax doesn't allow `args.from`.

    # Set appropriate log level
    log_format = "%(levelname)s (%(name)s) - %(message)s"
    if args.debug:
        logging.basicConfig(level=logging.DEBUG,
                            format=log_format)
    elif args.info:
        logging.basicConfig(level=logging.INFO,
                            format=log_format)
    else:
        logging.basicConfig(format=log_format)
    logger.info("Logging enabled")

    # Identify location of repository
    repo = args.repo or os.getcwd()
    logger.info("Repository is: {}".format(repo))

    # Identify VCS
    vcs_handler = _identify_vcs(repo)
    if vcs_handler is None:
        logger.error("Failed to identify VCS! Are you running Autobump in the root of the repository?")
        exit(1)
    logger.info("VCS handler is {}".format(vcs_handler.__name__))
    vcs_get_commit = partial(vcs_handler.get_commit, repo)
    vcs_all_tags = partial(vcs_handler.all_tags, repo)
    vcs_last_tag = partial(vcs_handler.last_tag, repo)
    vcs_last_commit = partial(vcs_handler.last_commit, repo)

    # Identify language
    handler_map = {
        "py": python,
        "python": python,
        "java_ast": java_ast,
        "java_native": java_native,
        "clojure": clojure
    }
    lang_handler = handler_map.get(args.handler, None)
    if lang_handler is None:
        logger.error("Invalid handler {} specified!".format(args.handler))
        exit(1)
    logger.info("Language handler is {}".format(lang_handler.__name__))

    # Check for evaluation mode
    if args.evaluate:
        if not args.f or not args.to:
            logger.error("Evaluation mode requires supplying a range")
            exit(1)
        _evaluate(args, vcs_all_tags())
        exit(0)

    # Identify revisions
    try:
        a_revision = args.f or vcs_last_tag()
        logger.info("Earlier revision identified as {}".format(a_revision))
        b_revision = args.to or vcs_last_commit()
        logger.info("Later revision identified as {}".format(b_revision))
    except VersionControlException:
        logger.error("Failed to automatically determine comparison range.")
        exit(1)

    # Identify changelog policy
    changelog_file = None
    if args.changelog and not args.changelog_stdout:
        changelog_file = open(args.changelog, "w")
        logger.info("Writing changelog to {}".format(args.changelog))
    elif args.changelog_stdout and not args.changelog:
        changelog_file = sys.stdout
        logger.info("Writing changelog to stdout")
    elif args.changelog and args.changelog_stdout:
        logger.error("`--changelog` and `--changelog-stdout` are mutually exclusive")
        exit(1)

    # Determine bump
    a_handle, a_location = vcs_get_commit(a_revision)
    b_handle, b_location = vcs_get_commit(b_revision)
    if lang_handler.build_required:
        logger.info("Handler indicated that a build is required")
        # Options "--build-command" and "--build-root" should be passed in.
        if not args.build_command or not args.build_root:
            logger.error("The {} handler requires that the project is built, but no build command or build root were provided".format(args.handler))
            exit(1)
        a_units = lang_handler.codebase_to_units(a_location, args.build_command, args.build_root)
        b_units = lang_handler.codebase_to_units(b_location, args.build_command, args.build_root)
        # Need to set the 'location' property of all types in both codebases
        # to the location of the latter one. Comparing types may require
        # loading compiled components.
        b_build_location = os.path.join(b_location, args.build_root)
        _patch_types_with_location(a_units, b_build_location)
        _patch_types_with_location(b_units, b_build_location)
    else:
        logger.info("Handler indicated no build is required")
        if args.build_command or args.build_root:
            logger.warn("No build is required, but build-command or build-root given - IGNORING")
        a_units = lang_handler.codebase_to_units(a_location)
        b_units = lang_handler.codebase_to_units(b_location)

    logger.debug("Found {} units in variant A".format(len(a_units)))
    logger.debug("Found {} units in variant B".format(len(b_units)))
    bump = diff.compare_codebases(a_units, b_units, changelog_file)
    logger.info("Bump found to be {}".format(bump))
    if changelog_file not in {None, sys.stdout}:
        changelog_file.close()
        logger.debug("Changelog file closed")

    # Determine version
    a_version = Semver.from_string(args.from_version) if args.from_version is not None else Semver.guess_from_string(a_revision)
    logger.debug("Earlier version is {}".format(a_version))
    b_version = a_version.bump(bump)
    logger.debug("Later version is {}".format(b_version))

    # Clean up temporary directories
    a_handle.cleanup()
    b_handle.cleanup()

    # Return the new version if autobump is called from another Python program.
    # If called from an entry point as specified in setup.py, this is printed.
    logger.info("Final answer, new version is {}".format(b_version))
    return b_version


def main(_):
    # Do not run anything if the module is just imported.
    pass



# Copyright 2016-2017 Christian Shtarkov
#
# This file is part of Autobump.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Codebase diffing logic.

This is the heart of Autobump.

The main entrypoint to this module 'compare_codebases'
is used to determine what the appropriate version bump should be
by considering how a codebase has changed over time.

Those changes can also be emmitted to a changelog file if requested.
"""

import logging
import functools
from enum import Enum

from autobump import config
from autobump.capir import Unit

logger = logging.getLogger(__name__)


@functools.total_ordering
class Bump(Enum):
    none = 0
    patch = 1
    minor = 2
    major = 3

    def __lt__(self, other):
        if type(self) is type(other):
            return self.value < other.value
        raise NotImplementedError()


class Change(Enum):
    removed_default_value = "Removed a default value"
    changed_default_value = "Changed a default value"
    entity_was_introduced = "Entity was introduced"
    entity_was_removed = "Entity was removed"
    function_was_overloaded = "Function was overloaded"
    overloaded_function_removed = "Overloaded function removed"
    parameter_added_to_signature = "Parameter(s) added to function signature"
    parameter_defaults_added_to_signature = "Parameter(s) with default value(s) added to function signature"
    parameter_removed_from_signature = "Parameter(s) removed from function signature"
    type_changed_to_compatible_type = "Type was changed to a compatible type"
    type_changed_to_incompatible_type = "Type was changed to an incompatible type"

    @staticmethod
    def get_bump(change):
        bump_map = {
            Change.removed_default_value: Bump.major,
            Change.changed_default_value: Bump.patch,
            Change.entity_was_introduced: Bump.minor,
            Change.entity_was_removed: Bump.major,
            Change.function_was_overloaded: Bump.minor,
            Change.overloaded_function_removed: Bump.major,
            Change.parameter_added_to_signature: Bump.major,
            Change.parameter_defaults_added_to_signature: Bump.minor,
            Change.parameter_removed_from_signature: Bump.major,
            Change.type_changed_to_compatible_type: Bump.patch,
            Change.type_changed_to_incompatible_type: Bump.major
        }
        return bump_map.get(change)


def _join_path(path, name):
    """Join a prefix path and an entity name to form a new path."""
    return (path + "." if path != "" else "") + name


def _compare_types(a_ent, b_ent):
    """Compare types of two entities and return a list of Changes."""
    changes = []
    if a_ent.type != b_ent.type:
        logger.debug("Types are different for {} and {}:\n\tVariant A: {}\n\tVariant B: {}"
                     .format(a_ent, b_ent, a_ent.type, b_ent.type))
        if b_ent.type.is_compatible(a_ent.type):
            logger.debug("Furthermore, types are compatible")
            changes.append(Change.type_changed_to_compatible_type)
        else:
            logger.debug("Furthermore, types are NOT compatible")
            changes.append(Change.type_changed_to_incompatible_type)
    return changes


def _compare_signatures(a_ent, b_ent):
    """Compare signatures of two entities and return a list of Changes."""
    changes = []
    logger.debug("Comparing signatures of {} and {}\n\tVariant A: {}\n\tVariant B: {}"
                 .format(a_ent, b_ent, str(a_ent.signatures), str(b_ent.signatures)))

    if len(a_ent.signatures) == 1 and len(b_ent.signatures) == 1:
        return _compare_signatures_directly(a_ent.signatures[0], b_ent.signatures[0])
    if len(a_ent.signatures) < len(b_ent.signatures):
        logger.debug("Variant A has less signatures")
        changes.append(Change.function_was_overloaded)
    elif len(a_ent.signatures) > len(b_ent.signatures):
        logger.debug("Variant B has less signatures")
        changes.append(Change.overloaded_function_removed)

    a_signatures = set(a_ent.signatures)
    b_signatures = set(b_ent.signatures)
    not_in_b = a_signatures.difference(b_signatures)
    not_in_a = b_signatures.difference(a_signatures)
    compat_signatures = set()
    for a_sig in not_in_b:
        for b_sig in not_in_a:
            sig_changes = _compare_signatures_directly(a_sig, b_sig)
            sig_changes = [c for c in sig_changes if Change.get_bump(c) > Bump.patch]
            if len(sig_changes) == 0:
                compat_signatures.add(a_sig)
                compat_signatures.add(b_sig)
    if len(compat_signatures) > 0:
        logger.debug("There were some different, but compatible signatures\n\t{}"
                     .format(str(compat_signatures)))
    not_in_a = not_in_a.difference(compat_signatures)
    not_in_b = not_in_b.difference(compat_signatures)
    for signature in not_in_a:
        logger.debug("Signature added in variant B")
        changes.append(Change.function_was_overloaded)
    for signature in not_in_b:
        logger.debug("Signature missing from variant B")
        changes.append(Change.overloaded_function_removed)

    return changes


def _compare_signatures_directly(a_signature, b_signature):
    """Compare two Signature objects directly and return a list of Changes."""
    changes = []
    a_parameters = a_signature.parameters
    b_parameters = b_signature.parameters
    logger.debug("Comparing signatures directly")

    # Check for type compatibility
    for pi in range(min(len(a_parameters), len(b_parameters))):
        changes = changes + _compare_types(a_parameters[pi], b_parameters[pi])

    # Check whether size of signature has changed
    if len(a_parameters) < len(b_parameters):
        # Signature was expanded - check for default values.
        logger.debug("Signature was expanded in later version: {} became {}"
                     .format(len(a_parameters), len(b_parameters)))
        all_new_have_defaults = True
        for pi in range(len(a_parameters), len(b_parameters)):
            if b_parameters[pi].default_value is None:
                logger.debug("At least one new parameter missing default value: {}"
                             .format(b_parameters[pi]))
                all_new_have_defaults = False
                break
        if all_new_have_defaults:
            logger.debug("All new parameters have default values")
            changes.append(Change.parameter_defaults_added_to_signature)
        else:
            logger.debug("NOT all new parameters have default values")
            changes.append(Change.parameter_added_to_signature)
    elif len(a_parameters) > len(b_parameters):
        # Signature has shrunk - always a breaking change.
        logger.debug("Signature has shrunk in later version: {} became {}"
                     .format(len(a_parameters), len(b_parameters)))
        changes.append(Change.parameter_removed_from_signature)

    return changes


def _compare_entities(a_ent, b_ent, changelog_file, path=""):
    """Compare two code entities which have the same name.

    Return a Bump enum based on whether
    there was a major, minor, patch or no change.
    """
    assert a_ent.name == b_ent.name, "Shouldn't compare entities with different names."
    assert type(a_ent) is type(b_ent), "Shouldn't compare entities of different types."

    path = _join_path(path, a_ent.name)
    logger.debug("Comparing {}" .format(path))

    if config.entity_ignored(path) and a_ent.name != "":
        logger.debug("Ignoring because of configuration")
        return Bump.patch

    highestBump = Bump.patch  # Biggest bump encountered so far.

    def _report_change(change, path):
        if changelog_file is not None:
            print("{}: {}".format(path, change.value), file=changelog_file)
        _report_bump(Change.get_bump(change))

    def _report_bump(bump):
        nonlocal highestBump
        if bump.value > highestBump.value:
            highestBump = bump

    # Map of the form:
    # (attribute required) -> (comparison function)
    comparisons = {
        "type": _compare_types,
        "signatures": _compare_signatures
    }

    for attribute, comparator in comparisons.items():
        if hasattr(a_ent, attribute):
            assert hasattr(b_ent, attribute), "Should never happen: entities have mismatching attributes."
            for change in comparator(a_ent, b_ent):
                _report_change(change, path)

    # Compare inner entities recursively
    for k, v in a_ent.__dict__.items():
        if type(v) is not dict:
            continue
        assert k in b_ent.__dict__, "Should never happen: comparing entities with different inner entities."
        a_inner = a_ent.__dict__[k]
        b_inner = b_ent.__dict__[k]
        for ki in {**a_inner, **b_inner}:

            if ki not in a_inner:
                # Handle case when a entity was added.
                logger.debug("Not found in variant A: {}".format(b_inner[ki].name))
                _report_change(Change.entity_was_introduced, _join_path(path, b_inner[ki].name))
                continue

            if ki not in b_inner:
                # Handle case when a entity was removed.
                logger.debug("Not found in variant B: {}".format(a_inner[ki].name))
                _report_change(Change.entity_was_removed, _join_path(path, a_inner[ki].name))
                continue

            # Handle general case when a entity may have changed.
            _report_bump(_compare_entities(a_inner[ki], b_inner[ki], changelog_file, path))

    return highestBump


def compare_codebases(a_units, b_units, changelog_file):
    """Compare codebases consisting of Units.

    Return a Bump enum based on whether
    there was a major, minor, patch or no change."""
    # Represent both codebases as a single unit, and compare that.
    a_unit = Unit("", dict(), dict(), a_units)
    b_unit = Unit("", dict(), dict(), b_units)
    return _compare_entities(a_unit, b_unit, changelog_file)



# Copyright 2016-2017 Christian Shtarkov
#
# This file is part of Autobump.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Convert a Python codebase into a list of Units.
"""

import os
import ast
import sys
import codecs
import logging
import traceback

from autobump import config
from autobump.capir import Type, Field, Parameter, Signature, Function, Unit

logger = logging.getLogger(__name__)

_source_file_ext = ".py"


class _PythonType(Type):
    pass


class _Dynamic(_PythonType):
    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "dynamic"


class _StructuralType(_PythonType):
    def __init__(self, attr_set):
        self.name = str(attr_set)
        self.attr_set = attr_set

    def is_compatible(self, other):
        if not isinstance(other, _StructuralType):
            return False
        return self.attr_set.issubset(other.attr_set)

    def __str__(self):
        return str(self.attr_set)


class _HintedType(_PythonType):
    def __init__(self, name):
        self.name = name

    def is_compatible(self, other):
        return self.__eq__(other)

    def __eq__(self, other):
        return self.name == other.name

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.__str__()


_dynamic = _Dynamic()


def _is_public(member_name):
    """Determine visibility of a member based on its name."""
    return not (member_name.startswith("_") and member_name != "__init__")


def _get_type_of_parameter(function, parameter):
    """Return the type of a parameter used in a function AST node.

    In this case, 'type' means structural instead of nominal type.
    Because Python is dynamically typed, it would be very hard to guess
    what type a parameter is without looking at usage. Instead of doing that,
    this walks the AST node describing the function and considers the type to be
    the set of all methods called on the parameter."""
    assert isinstance(function, ast.FunctionDef), "Tried to get usage of parameter in a non-function."

    # Check if there is a type hint for this parameter
    if config.type_hinting():
        for arg in function.args.args:
            if arg.arg == parameter:
                if arg.annotation:
                    return _HintedType(arg.annotation.id)

    if not config.structural_typing():
        return _dynamic

    # TODO: Don't completely ommit 'self' in class methods,
    # it can be used to identify addition or removal of fields.
    if parameter == "self":
        return _StructuralType(set())

    # Generators to filter out AST
    def gen_no_inner_definitions(node):
        """Recursively yield all descendant nodes
        without walking any function or class definitions."""
        yield node
        for n in ast.iter_child_nodes(node):
            if isinstance(n, ast.FunctionDef) or \
               isinstance(n, ast.ClassDef):
                continue
            yield from gen_no_inner_definitions(n)

    def gen_only_attributes(node):
        """Yield only descendant nodes that represent attribute access,
        without traversing any function or class definitions."""
        for n in gen_no_inner_definitions(node):
            if isinstance(n, ast.Attribute) and \
               isinstance(n.value, ast.Name):
                yield n

    # Find the set of attributes for that parameter
    attr_set = set()
    for attr in gen_only_attributes(function):
        name = attr.value.id
        method = attr.attr
        if name == parameter:
            # TODO: Also consider method signature.
            attr_set.add(method)

    # Convert set of attribytes to structural type
    return _StructuralType(attr_set)


def _get_signature(function):
    """Return the signature of a function AST node."""
    parameters = []
    args = function.args.args

    # Map all None parameters to a "TrueNone" object
    # because None indicates the absense of a default value.
    class TrueNone(object):
        pass
    defaults = [TrueNone
                if isinstance(a, ast.NameConstant) and a.value is None
                else a
                for a in function.args.defaults]
    # Prepend no default values.
    defaults = [None] * (len(args) - len(defaults)) + defaults

    args_with_defaults = list(zip(args, defaults))
    for arg_with_default in args_with_defaults:
        arg, default = arg_with_default
        if isinstance(default, ast.Name):
            # TODO: This does not differentiate between
            # "abc" and abc.
            default = default.id
        elif isinstance(default, ast.NameConstant):
            default = default.value
        elif isinstance(default, ast.Num):
            default = default.n
        elif isinstance(default, ast.Str):
            default = default.s
        type = _get_type_of_parameter(function, arg.arg)
        parameters.append(Parameter(arg.arg, type, default))
    # Note: we need to return a list with the signature inside
    # because the common representation allows for overloading,
    # which Python doesn't.
    return [Signature(parameters)]


def _container_to_unit(name, container):
    """Convert a Python AST module or class to a Unit."""
    fields = dict()
    functions = dict()
    units = dict()
    for node in container.body:
        if hasattr(node, "name") and not _is_public(node.name):
            # Completely ignore any private things -
            # they are irrelevant to the API.
            continue
        if isinstance(node, ast.ClassDef):
            units[node.name] = _container_to_unit(node.name, node)
        elif isinstance(node, ast.FunctionDef):
            functions[node.name] = Function(node.name, _dynamic, _get_signature(node))
        elif isinstance(node, ast.Assign):
            # TODO: Handle other forms of assignment.
            for target in [t for t in node.targets if isinstance(t, ast.Name) and _is_public(t.id)]:
                fields[target.id] = Field(target.id, _dynamic)
    return Unit(name, fields, functions, units)


def _module_to_unit(name, module):
    """Convert a Python AST module to a Unit."""
    return _container_to_unit(name, module)


def python_codebase_to_units(location):
    """Returns a list of Units representing a Python codebase in 'location'."""
    if config.type_hinting():
        # When the handler is invoked, the 'ast' module needs to start
        # pointing to 'ast35' from 'typed_ast' if type hinting is to be used.
        # Note that 'ast' must be changed globally, as the other functions in this
        # module rely on it as well.
        global ast
        from typed_ast import ast35
        ast = ast35

    units = dict()
    for root, dirs, files in os.walk(location):
        dirs[:] = [d for d in dirs if not config.dir_ignored(d)]
        pyfiles = [f for f in files if f.endswith(_source_file_ext) and not config.file_ignored(f)]
        for pyfile in pyfiles:
            pymodule = pyfile[:-(len(_source_file_ext))]  # Strip extension
            with codecs.open(os.path.join(root, pyfile),
                             "r",
                             encoding="utf-8",
                             errors="replace") as f:
                try:
                    units[pymodule] = _module_to_unit(pymodule, ast.parse(f.read()))
                except Exception:
                    print(traceback.format_exc(), file=sys.stderr)
                    msg = "Failed to parse file {}".format(os.path.join(root, pyfile))
                    if config.python_omit_on_error():
                        logger.warning(msg)
                    else:
                        logger.error(msg)
                        exit(1)

    return units

build_required = False
codebase_to_units = python_codebase_to_units



import sys
from setuptools import setup

if sys.version_info < (3, 5):
    sys.exit("Python 3.5+ is required for Autobump.")

setup(name="autobump",
      version="0.2.0",
      description="Automatic semantic versioning of projects",
      url="https://github.com/cshtarkov/autobump",
      author="Christian Shtarkov",
      author_email="cshtarkov@gmail.com",
      license="GPLv3",
      packages=["autobump", "autobump.handlers"],
      package_dir={"autobump": "autobump"},
      package_data={"autobump": ["libexec/*"]},
      install_requires=["javalang", "typed-ast"],
      entry_points={
          "console_scripts": ["autobump = autobump:autobump"]
      })



# Copyright 2016-2017 Christian Shtarkov
#
# This file is part of Autobump.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Convert a Java codebase into a list of Units using the AST.

This handler does not require any external utilities. It uses the library
'javalang' to get the AST of all Java files in the project.
"""

import os
import copy
import logging
import javalang

from autobump import config
from autobump.capir import Type, Field, Parameter, Signature, Function, Unit

logger = logging.getLogger(__name__)

_primitive_types = {"void", "byte", "short", "int", "long", "float", "double", "boolean", "char"}
_source_file_ext = ".java"


class _JavaType(Type):

    def __init__(self, name, dimension=0):
        self.name = name
        self.dimension = dimension
        self.children = set()

    def is_compatible(self, other):
        # Check if it's the same type
        if self == other:
            return True
        # Check if it's the same dimension.
        if self.dimension != other.dimension:
            return False
        # Check if it's a direct descendant.
        if len({jt for jt in self.children if jt.name == other.name}) == 1:
            return True
        # Recursively check all children.
        for child in self.children:
            if child.is_compatible(other):
                return True
        return False

    def __eq__(self, other):
        return self.name == other.name and self.dimension == other.dimension

    def __hash__(self):
        return hash((self.dimension, self.name))

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<JavaType {} {}, children: {}>".format(self.dimension, self.name, self.children)


class _JavaTypeSystem(object):
    """Represents a collection of types available to a Java codebase.

    Use 'add_qualified_type' to add fully-qualified type names to the system,
    e.g. "java.util.List".

    When done, use 'finalize' to resolve all types and build Type
    objects. Type objects can be looked up by using 'lookup'."""

    def __init__(self):
        self.types = dict()
        self.finalized = False

    def add_qualified_type(self, type_name, type_node_and_compilation):
        """Add a fully qualified 'type_name' with its AST node and compilation unit."""
        assert type(type_node_and_compilation) is tuple
        # Add the type as it is
        self.types[type_name] = type_node_and_compilation

    def finalize(self):
        """Replace all Type AST nodes with Type objects."""
        self.type_objects = {}
        for type_name in self.types.keys():
            node, compilation = self.types[type_name]
            if type_name not in self.type_objects:
                self.type_objects[type_name] = _JavaType(type_name)
            type_object = self.type_objects[type_name]

            # Check 'extends' for this type
            if hasattr(node, "extends") and node.extends is not None:
                extends = _qualify_type(node.extends, compilation)
                if extends in self.type_objects:
                    self.type_objects[extends].children.add(type_object)
                else:
                    extended_object = _JavaType(extends)
                    self.type_objects[extends] = extended_object
                    self.type_objects[extends].children.add(type_object)

            # Check 'implements' for this type
            if hasattr(node, "implements") and node.implements is not None:
                implements = [_qualify_type(r, compilation) for r in node.implements]
                for interface in implements:
                    if interface in self.type_objects:
                        self.type_objects[interface].children.add(type_object)
                    else:
                        interface_object = _JavaType(interface)
                        self.type_objects[interface] = interface_object
                        self.type_objects[interface].children.add(type_object)

        # Also need to add all primitive types
        for primitive_type in _primitive_types:
            type_object = _JavaType(primitive_type)
            self.type_objects[primitive_type] = type_object

        self.types = self.type_objects
        self.finalized = True

    def lookup(self, type_name):
        """Lookup a fully-qualified type name and get a Type object.

        Implicitly finalizes the JavaTypeSystem, if not done already."""
        if not self.finalized:
            self.finalize()

        dimension = _array_dimension(type_name)
        type_wo_dimension = _strip_array_dimension(type_name)
        if type_wo_dimension not in self.types:
            message = "{} is an external type".format(type_wo_dimension)
            if config.java_error_on_external_types():
                logger.error(message)
                exit(1)
            else:
                logger.warning(message)
                return _JavaType(type_name)

        type_object = copy.deepcopy(self.types[type_wo_dimension])
        type_object.dimension = dimension
        return type_object

    def qualify_lookup(self, type, compilation):
        """Qualify and lookup a type name and get a Type object.

        Implicitly finalizes the JavaTypeSystem, if not done already."""
        return self.lookup(_qualify_type(type, compilation))


_dummyType = _JavaType("dummy")
_dummyType.is_compatible = lambda t: True


def _array_dimension(name):
    for dimension in range(len(name)):
        if name[dimension] != '[':
            break
    return dimension


def _prefix_array_dimension(name, dimension):
    return "[" * dimension + name


def _strip_array_dimension(name):
    return name[_array_dimension(name):]


def _is_public(node):
    """Determine visibility of an AST node."""
    return hasattr(node, "modifiers") and "public" in node.modifiers


def _get_declarator_names(field):
    """Return list of declarator names of a field AST node."""
    assert isinstance(field, javalang.tree.FieldDeclaration)
    return [d.name for d in field.declarators]


def _qualify_type(type, compilation):
    """Return the fully-qualified name of a type reference in a compilation unit."""
    name = type.name
    dimension = len(getattr(type, "dimensions", []))
    if name in _primitive_types:
        return _prefix_array_dimension(name, dimension)
    candidates = [i for i in compilation.imports
                  if i.path.endswith(name) and (i.path.find(name) == 0 or i.path[i.path.find(name) - 1] == ".")]
    assert len(candidates) <= 1, \
           "Don't know what to do: more than one candidate for qualifying {} found in {}".format(name, compilation.filename)
    if len(candidates) == 1:
        # Import statement exactly matched - must be this definition.
        return _prefix_array_dimension(candidates[0].path, dimension)
    else:
        # No import statement matched - should be in this package.
        # TODO: Handle wildcard imports somehow.
        assert name.find(".") == -1, \
               "Should never happen: no import statement found, yet {} is relative!".format(name)
        if compilation.package is not None:
            return _prefix_array_dimension(compilation.package.name + "." + name, dimension)
        else:
            return _prefix_array_dimension(name, dimension)


def _get_parameters(method, type_system, compilation):
    """Return a list of Parameters to a method."""
    parameters = []
    for p in method.parameters:
        name = p.name
        type = type_system.qualify_lookup(p.type, compilation)
        parameters.append(Parameter(name, type))
    return parameters


def _get_return_type(method, type_system, compilation):
    """Return the return type of a method."""
    if method.return_type is None:
        void = javalang.tree.Type(name="void", dimensions=[])
        return type_system.qualify_lookup(void, compilation)
    return type_system.qualify_lookup(method.return_type, compilation)


def _class_or_interface_to_unit(node, compilation, type_system):
    """Convert a class or interface declaration ('node') into a Unit.

    Requires the 'compilation' where the node so it can be used to look up type information."""
    assert isinstance(node, javalang.tree.ClassDeclaration) or \
           isinstance(node, javalang.tree.InterfaceDeclaration) or \
           isinstance(node, javalang.tree.AnnotationDeclaration) or \
           isinstance(node, javalang.tree.EnumDeclaration)

    # TODO: Handle annotation declarations
    if isinstance(node, javalang.tree.AnnotationDeclaration):
        return Unit("annotation", {}, {}, {})
    # TODO: Handle enum declarations
    if isinstance(node, javalang.tree.EnumDeclaration):
        return Unit("enum", {}, {}, {})

    fields = dict()
    functions = dict()
    units = dict()
    for n in [n for n in node.body if _is_public(n)]:

        # Convert fields.
        if isinstance(n, javalang.tree.FieldDeclaration):
            for declarator_name in _get_declarator_names(n):
                type_object = type_system.qualify_lookup(n.type, compilation)
                fields[declarator_name] = Field(declarator_name, type_object)

        # Convert methods.
        elif isinstance(n, javalang.tree.MethodDeclaration):
            parameters = _get_parameters(n, type_system, compilation)
            return_type = _get_return_type(n, type_system, compilation)
            # Java supports method overloading, where methods with the
            # same name can be differentiated by their parameters and return type.
            # To fit this into autobump's common representation, we set
            # the first parameter of the method to be the return type.
            # Then, we set a dummy type that is compatible with anything as the
            # actual return type of the function.
            parameters = [Parameter("$AUTOBUMP_RETURN$", return_type)] + parameters
            if n.name in functions:
                functions[n.name].signatures.append(Signature(parameters))
            else:
                functions[n.name] = Function(n.name,
                                                    _dummyType,
                                                    [Signature(parameters)])

        # Convert inner classes and interfaces.
        elif isinstance(n, javalang.tree.ClassDeclaration) or \
             isinstance(n, javalang.tree.InterfaceDeclaration):
            units[n.name] = _class_or_interface_to_unit(n, compilation, type_system)

    fqn = _qualify_type(node, compilation)
    return Unit(fqn, fields, functions, units)


def _source_to_compilation(filename, source):
    """Convert source text from 'filename' into a compilation unit."""
    try:
        tree = javalang.parse.parse(source)
    except javalang.parser.JavaSyntaxError as e:
        logger.error("Java Syntax Error  {}:{}: {}"
                     .format(filename, e.at, e.description))
        logger.error("Stopped parsing {}".format(filename))
        exit(1)
    tree.filename = filename
    return tree


def _compilation_to_unit(compilation, type_system):
    """Convert a compilation unit to a Unit.

    In Java, there can be at most one class or interface declaration
    per file. To avoid redundancy, the file is not considered a separate
    Unit."""
    assert len(compilation.types) == 1, \
           "Don't know what to do: {} contains more than one top-level type declaration!".format(compilation.filename)
    return _class_or_interface_to_unit(compilation.types[0], compilation, type_system)


def _compilation_get_types(compilation):
    """Get all types declared in compilation unit."""
    types = []
    prefix = compilation.package.name + "." if compilation.package is not None else ""

    def get_inner_definitions(node, path):
        types = []
        for n in node.body:
            if not (isinstance(n, javalang.tree.ClassDeclaration) or
                    isinstance(n, javalang.tree.InterfaceDeclaration)):
                continue
            types.append((path + n.name, n))
            types = types + get_inner_definitions(n, path + n.name + ".")
        return types

    for type in compilation.types:
        types.append((prefix + type.name, type))
        types = types + get_inner_definitions(type, prefix + type.name + ".")

    return types


def java_codebase_to_units(location):
    """Return a list of Units representing a Java codebase in 'location'."""

    # This needs to perform two passes on all Java files.
    #
    # In the first pass:
    #   1) Java files are translated to compilation units (aka AST).
    #   2) The type system is enriched with all types declared in this codebase.
    #
    # In the second pass:
    #   1) Referential types are resolved when referring to this codebase.
    #   2) All compilation units are translated into Units.

    units = dict()

    # First pass
    type_system = _JavaTypeSystem()
    compilations = []
    for root, dirs, files in os.walk(location):
        dirs[:] = [d for d in dirs if not config.dir_ignored(d)]
        javafiles = [f for f in files if f.endswith(_source_file_ext) and not config.file_ignored(f)]
        for javafile in javafiles:
            with open(os.path.join(root, javafile), "r") as f:
                compilation = _source_to_compilation(javafile, f.read())
                for type_name, type_node in _compilation_get_types(compilation):
                    type_system.add_qualified_type(type_name, (type_node, compilation))
                compilations.append(compilation)

    type_system.finalize()

    # Second pass
    for compilation in compilations:
        unit = _compilation_to_unit(compilation, type_system)
        units[unit.name] = unit

    return units

build_required = False
codebase_to_units = java_codebase_to_units
